import tkinter as tk
import numpy as np
import QTM_Core as Cor
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg)
import os
import shutil
import QTM_TT_IO as TT
from tkinter import filedialog as fd
from scipy import signal
from scipy import stats
from scipy.ndimage import gaussian_filter

def Calc_PSD(itSlot):
    strt = int(max(itSlot,2)); stop = int(min(itSlot + Cor.NoiseSlotWidth,len(NoiseDfDGbyns)-2))
    amps    = np.ones(Cor.novt)*np.nan
    PLExpts = np.ones(Cor.novt)*np.nan
    for count,iovt in enumerate(Cor.iovts_inc):
        xs = np.linspace(0,stop-strt,stop-strt,endpoint=False)
        ys = NoiseDfDGbyns[strt:stop,2+4*iovt]
        if Cor.NoiseCompParablDrift == 'no':
            p_fitted = np.polyfit(xs, ys, deg=1)
            ys -= p_fitted[0] + p_fitted[1] * xs
        if Cor.NoiseCompParablDrift == 'yes':
            p_fitted = np.polyfit(xs, ys, deg=2)
            ys -= p_fitted[0] + p_fitted[1] * xs + p_fitted[2] * xs**2
        fs,PSD_Dfbyn = signal.welch(ys,nperseg = len(ys))

        ys = NoiseDfDGbyns[strt:stop,2+4*iovt+1]
        if Cor.NoiseCompParablDrift == 'no':
            p_fitted = np.polyfit(xs, ys, deg=1)
            ys -= p_fitted[0] + p_fitted[1] * xs
        if Cor.NoiseCompParablDrift == 'yes':
            p_fitted = np.polyfit(xs, ys, deg=2)
            ys -= p_fitted[0] + p_fitted[1] * xs + p_fitted[2] * xs**2
        fs,PSD_DGbyn = signal.welch(ys,nperseg = len(ys))
                
        PSD = 0.5*(PSD_Dfbyn+PSD_DGbyn) 
        PSD = PSD[1:]; fs  = fs[1:]
        if count == 0 : PSDs = [PSD]
        else          : PSDs = np.append(PSDs,[PSD],axis=0)
        result = stats.linregress(np.log(fs),np.log(PSD))
        amps[iovt]    = np.exp(result.intercept)
        PLExpts[iovt] = result.slope
        
    return fs,PSDs,PLExpts,amps

def Calc_rms_noises():
    global times,Hadamard,rms_noises
    strt = int(max(Cor.i_strt_fit,2)); 
    stop = int(min(Cor.i_stop_fit,len(NoiseDfDGbyns)-2)); 
    times = np.ones((stop-strt))*np.nan
    Hadamard   = np.ones((Cor.novt,stop-strt))*np.nan
    rms_noises = np.ones((Cor.novt,stop-strt))*np.nan
    for iovt_inc in Cor.iovts_inc:
        count = 0
        for it in range(strt,stop):
            times[count] = NoiseDfDGbyns[it,1]
            if Cor.Noise_Dfbyn_DGbyn_both == 'Dfbyn': list_Dfbyn_DGbyn_both = [0]
            if Cor.Noise_Dfbyn_DGbyn_both == 'DGbyn': list_Dfbyn_DGbyn_both = [1]
            if Cor.Noise_Dfbyn_DGbyn_both == 'both' : list_Dfbyn_DGbyn_both = [0,1]
            Hadamard[iovt_inc,count] = 0
            count2 = 0
            for iDfDG in list_Dfbyn_DGbyn_both:
                if Cor.NoiseCompParablDrift == 'no':
                    Hadamard[iovt_inc,count] += \
                         1./6. *(NoiseDfDGbyns[it-1,2+4*iovt_inc+iDfDG]\
                          - 2.* NoiseDfDGbyns[it  ,2+4*iovt_inc+iDfDG]\
                              + NoiseDfDGbyns[it+1,2+4*iovt_inc+iDfDG])**2
                if Cor.NoiseCompParablDrift == 'yes':
                    Hadamard[iovt_inc,count] += \
                       1./70. *(NoiseDfDGbyns[it-2,2+4*iovt_inc+iDfDG]\
                          - 4.* NoiseDfDGbyns[it-1,2+4*iovt_inc+iDfDG]\
                          + 6.* NoiseDfDGbyns[it  ,2+4*iovt_inc+iDfDG]\
                          - 4.* NoiseDfDGbyns[it+1,2+4*iovt_inc+iDfDG]\
                              + NoiseDfDGbyns[it+2,2+4*iovt_inc+iDfDG])**2
                count2 += 1
            Hadamard[iovt_inc,count] /= count2    
            rms_noises[iovt_inc,count] = Hadamard[iovt_inc,count]**0.5
            count +=1
        rms_noises[iovt_inc] = gaussian_filter(rms_noises[iovt_inc], sigma=Cor.NoiseSmoothWidth)
    
def AnalyzeNoise():
    global times,timeSlots,amps_all,PLExpts_all,fs,PSDs,PSDs_all,NoiseDfDGbyns
    NoiseDfDGbyns = np.copy(TT.DfDGbyns)
    Calc_rms_noises()
    timeSlots = []
    itSlots = []
    for it in range(len(NoiseDfDGbyns)):
        if (it-2)%Cor.NoiseSlotWidth == 0 and \
        it >= Cor.i_strt_fit and it <= Cor.i_stop_fit: 
            itSlots   = np.append(itSlots,it)
            timeSlots = np.append(timeSlots,NoiseDfDGbyns[it,1])
    itSlots   = itSlots[  :-1].astype(int)
    timeSlots = timeSlots[:-1]

    amps_all              = np.ones((Cor.novt,len(itSlots)))*np.nan
    PLExpts_all           = np.ones((Cor.novt,len(itSlots)))*np.nan
    for iSlot,itSlot in enumerate(itSlots):
        fs,PSDs,PLExpts_all[:,iSlot],amps_all[:,iSlot] = Calc_PSD(itSlot)
        if iSlot == 0 : PSDs_all = [PSDs]
        else          : PSDs_all = np.append(PSDs_all,[PSDs],axis=0)
    header = 'time \t '
    dataout = [timeSlots]
    for iovt in range(Cor.novt):
        dataout = np.append(dataout,[amps_all[iovt]])
        header = header + 'amps' + str(iovt) + '\t'
    for iovt in range(Cor.novt):
        dataout = np.append(dataout,[PLExpts_all[iovt]])
        header = header + 'PLExpts' + str(iovt) + '\t'
    np.savetxt(os.getcwd()+'/tmp/AnalyzeNoise.qt2',np.transpose(dataout),header=header,delimiter = '\t') 

def Plot_amps_PLExpts(contr):
    for widget in contr.winfo_children(): widget.destroy()
    fig = Figure(figsize = (4,6))
    axis = fig.add_subplot(311)
    for iovt_inc in Cor.iovts_inc:
        axis.plot(times,rms_noises[iovt_inc]/1e-3,'.',markersize = 2,\
                  color=Cor.Colors[iovt_inc])
    axis.set_ylabel('(Hadamar variance)^0.5\n [mHz]',fontsize=10)
    axis.tick_params(direction='in',labelsize =8)
    axis.axes.xaxis.set_ticklabels([])
    if np.nanmax(rms_noises) > 0 : axis.axes.set_yscale('log')

    axis = fig.add_subplot(312)
    for iovt_inc in Cor.iovts_inc:
        axis.plot(timeSlots,amps_all[iovt_inc]/1e-3,'.',markersize = 2.,\
                  color=Cor.Colors[iovt_inc])
    axis.set_ylabel('amplitudes [mHz]',fontsize=10)
    axis.tick_params(direction='in',labelsize =8)
    axis.axes.xaxis.set_ticklabels([])
    if np.nanmax(amps_all) > 0 : axis.axes.set_yscale('log')

    axis = fig.add_subplot(313)
    for iovt_inc in Cor.iovts_inc:
        axis.plot(timeSlots,PLExpts_all[iovt_inc],'.',markersize = 2.,\
                  color=Cor.Colors[iovt_inc])
    axis.tick_params(direction='in',labelsize=8)
    axis.set_xlabel('time ' + Cor.TimeLabel,fontsize=10)    
    axis.set_ylabel('Power Law Exponents',fontsize=10)

    suptitletext = ''
    if len(Cor.InfoString) > 0: 
        suptitletext += Cor.InfoString
    suptitletext +=  '\n' + os.path.basename(Cor.filename)[:55]
    fig.suptitle(suptitletext,fontsize = 8)
    fig.tight_layout(); 
    fig.savefig(os.getcwd()+'/tmp/AnalyzeNoise_amps_PLexpts.png',dpi=200)

    canvas = FigureCanvasTkAgg(fig,contr); canvas.draw(); 
    canvas.get_tk_widget().grid(row=1,column=0,columnspan=2,sticky='NW')

def Plot_PSDs(cntr):
    fig = Figure(figsize = (4,6))
    axis = fig.add_subplot(111)
    gap = np.max(np.log(PSDs_all[:,:,2:])) / 1000
    print('gap',gap)
    count = 0
    for iplot in range(0,len(timeSlots),int(np.max([len(timeSlots)/10,2]))):
        print('iplot',iplot)
        for iovt in range(len(PSDs_all[0])):
            axis.plot(np.log(fs[:-1]),np.log(PSDs_all[iplot,iovt,:-1]) + gap*count,'.-',markersize = 2.,
                      color = Cor.Colors[Cor.iovts_inc[iovt]])
        count += 1    
    axis.set_ylabel('Log PSDs',fontsize=10)
    axis.set_xlabel('Log Frequency ' + Cor.TimeLabel,fontsize=10)
    axis.tick_params(direction='in',labelsize =8)
    suptitletext = ''
    if len(Cor.InfoString) > 0: 
        suptitletext += Cor.InfoString
    suptitletext +=  '\n' + os.path.basename(Cor.filename)[:55]
    fig.suptitle(suptitletext,fontsize = 8)
    fig.tight_layout(); fig.savefig(os.getcwd()+'/tmp/AnalyzeNoise_amps_PLexpts.png',dpi=200)

    canvas = FigureCanvasTkAgg(fig,cntr); canvas.draw(); 
    canvas.get_tk_widget().grid(row=1,column=2,columnspan=7,sticky='NW')

def Export():
    ExportFilepath = fd.asksaveasfilename(initialfile='test_'+ os.path.basename(Cor.TT_filename[:-4]),\
            filetypes=[("qtm file", ".qtm")])
    if ExportFilepath == '' : return
    Folder          = os.path.dirname( ExportFilepath)
    ExportFilepath2 = os.path.basename(ExportFilepath).replace('_Master','')
    ExportFilepath2 = ExportFilepath2.replace('.qtm','')

    if os.path.isfile(  os.getcwd()+'/tmp/AnalyzeNoise.qt2'):
        shutil.copyfile(os.getcwd()+'/tmp/AnalyzeNoise.qt2',\
                        Folder + '/' + ExportFilepath2 + '_' + 'AnalyzeNoise.qt2')
    if os.path.isfile(  os.getcwd()+'/tmp/AnalyzeNoise_amps_PLExpts.png'):
        shutil.copyfile(os.getcwd()+'/tmp/AnalyzeNoise_amps_PLExpts.png',\
                        Folder + '/' + ExportFilepath2 + '_' + 'AnalyzeNoise_amps_PLExpts.png')
    if os.path.isfile(  os.getcwd()+'/tmp/AnalyzeNoise_PSDs.png'):
        shutil.copyfile(os.getcwd()+'/tmp/AnalyzeNoise_PSDs.png',\
                        Folder + '/' + ExportFilepath2 + '_' + 'AnalyzeNoise_PSDs.png')

def Make_AllFrame(cntr):
    for widget in cntr.winfo_children(): widget.destroy()
    Dfbyn_DGbyn_both_Labels = ['Df/n','DG/n','both']

    def Enter_Noise(event,WhichStuffStr,key):
        if WhichStuffStr == 'NoiseSlotWidth': 
            string = NoiseSlotWidth_SV.get(); string.strip()
            # try: 
            Cor.NoiseSlotWidth = int(string)
            Cor.NoiseSlotWidth = max(Cor.NoiseSlotWidth,8)
            Cor.Write_Config();
            Make_AllFrame(AllFrame); 
            # except: pass    
        if WhichStuffStr == 'NoiseSmoothWidth': 
            string = NoiseSmoothWidth_SV.get(); string.strip()
            # try: 
            Cor.NoiseSmoothWidth = int(string)
            Cor.Write_Config();
            Make_AllFrame(AllFrame); 
            # except: pass    
        
    def enter_Dfbyn_DGbyn_both(event):    
        Dfbyn_DGbyn_both = event.widget.get()
        for i,Dfbyn_DGbyn_both_label in enumerate(Dfbyn_DGbyn_both_Labels):
            if Dfbyn_DGbyn_both == Dfbyn_DGbyn_both_label: 
                Cor.Noise_Dfbyn_DGbyn_both = Cor.Noise_Dfbyn_DGbyn_both_strings[i]
        Cor.Write_Config()
        Make_AllFrame(AllFrame); 

    def Check_NoiseCompParablDrift(): 
        if int(NoiseCompParablDrift_IV.get()) == 1: Cor.NoiseCompParablDrift = 'yes'
        if int(NoiseCompParablDrift_IV.get()) == 0: Cor.NoiseCompParablDrift = 'no'
        Cor.Write_Config() 
        Make_AllFrame(AllFrame); 

    NoiseSlotWidth_SV = tk.StringVar(cntr,int(Cor.NoiseSlotWidth))
    NoiseSmoothWidth_SV = tk.StringVar(cntr,int(Cor.NoiseSmoothWidth))
    if Cor.NoiseCompParablDrift == 'yes': NoiseCompParablDrift_IV = tk.IntVar(cntr,1)
    if Cor.NoiseCompParablDrift == 'no' : NoiseCompParablDrift_IV = tk.IntVar(cntr,0)
    AnalyzeNoise()
    Plot_amps_PLExpts(cntr)
    Plot_PSDs(cntr)
    tk.Checkbutton(cntr,text='Compensate Parabolic Drift',variable=NoiseCompParablDrift_IV,\
        command=Check_NoiseCompParablDrift, anchor='w').grid(row=0,column=0,sticky='W')
    tk.Label(cntr,width=11,text='Slot Width',anchor='e').grid(row=0,column=1,sticky='E')
    NoiseSlotWidth_Entry = tk.Entry(cntr,width=5,textvariable=NoiseSlotWidth_SV      )
    NoiseSlotWidth_Entry.grid(row=0,column=2,sticky='W')
    NoiseSlotWidth_Entry.bind('<Return>',lambda event: Enter_Noise(event,'NoiseSlotWidth',''))
    tk.Label(cntr,text='Smooth Window Width',anchor='e').grid(row=0,column=3,sticky='E')
    NoiseSmoothWidth_Entry = tk.Entry(cntr,width=5,textvariable=NoiseSmoothWidth_SV      )
    NoiseSmoothWidth_Entry.grid(row=0,column=4,sticky='W')
    NoiseSmoothWidth_Entry.bind('<Return>',lambda event: Enter_Noise(event,'NoiseSmoothWidth',''))
    if Cor.Noise_Dfbyn_DGbyn_both == 'Dfbyn':
        Noise_Dfbyn_DGbyn_both_cb = tk.ttk.Combobox(cntr,state='readonly',width=12,\
            textvariable=Dfbyn_DGbyn_both_Labels[0],values=Dfbyn_DGbyn_both_Labels); 
        Noise_Dfbyn_DGbyn_both_cb.current(0)
    if Cor.Noise_Dfbyn_DGbyn_both == 'DGbyn':
        Noise_Dfbyn_DGbyn_both_cb = tk.ttk.Combobox(cntr,state='readonly',width=12,\
            textvariable=Dfbyn_DGbyn_both_Labels[1],values=Dfbyn_DGbyn_both_Labels); 
        Noise_Dfbyn_DGbyn_both_cb.current(1)
    if Cor.Noise_Dfbyn_DGbyn_both == 'both':
        Noise_Dfbyn_DGbyn_both_cb = tk.ttk.Combobox(cntr,state='readonly',width=12,\
            textvariable=Dfbyn_DGbyn_both_Labels[2],values=Dfbyn_DGbyn_both_Labels); 
        Noise_Dfbyn_DGbyn_both_cb.current(2)
    Noise_Dfbyn_DGbyn_both_cb.grid(row=0,column=5,sticky='W')
    Noise_Dfbyn_DGbyn_both_cb.bind('<<ComboboxSelected>>',enter_Dfbyn_DGbyn_both)

def AnalyzeNoise_Start():
    global AnalyzeNoiseRoot,AllFrame
    AnalyzeNoiseRoot = tk.Tk(); AnalyzeNoiseRoot.title('Analyze Noise')
    AllFrame = tk.LabelFrame(AnalyzeNoiseRoot); AllFrame.grid(row=0,column=0)
    Make_AllFrame(AllFrame)

    menubar = tk.Menu(AnalyzeNoiseRoot); AnalyzeNoiseRoot.config(menu=menubar) 
    menubar.add_command(label='Export',command=Export)
    
    AnalyzeNoiseRoot.iconbitmap("QTM.ico")
    AnalyzeNoiseRoot.mainloop()